#include "EventProgrammer.h"
#include "Constants.h"

static Uint8 _isInitialized = 0;
static Uint16 _frameMs;
static Uint32 _delayBetweenEventsMs;
static SDL_TimerID _timer;

#define DELAY_BETWEEN_EVENTS_IN_FRAMES 10   // 3 is too short, 4 begins to work
                                            // 6 sometimes fails in Debug mode
#define SECONDS_FROM_CPU_START_TO_COPYRIGHT_MESSAGE 2.5f

#define MAX_EVENTS 100
static SDL_Event _loadQuoteQuoteEvents[MAX_EVENTS];
static Uint16 _loadQuoteQuoteNumEvents;
static Uint16 _eventIndex;

Uint32 _process_event(void* param) {
    SDL_PushEvent(&_loadQuoteQuoteEvents[_eventIndex++]);

    if (_eventIndex == _loadQuoteQuoteNumEvents) {
        return 0;   // we're out of events, so we cancel timer
    }

    return _delayBetweenEventsMs;   // we delay until next event
}

void eventprogrammer_run_LOAD_QUOTE_QUOTE() {
    // in case we're already generating events
    SDL_RemoveTimer(_timer);

    _eventIndex = 0;

    double secondsAfterCPUStarts = (double)SECONDS_FROM_CPU_START_TO_COPYRIGHT_MESSAGE;
    Uint32 insertKeystrokesDelayMs = VIDEO_DELAY_TIMER_START_MS +                          // delay before CPU actually starts
        (Uint32)((secondsAfterCPUStarts * 1000.0f / (double)DEFAULT_FRAME_MS) * _frameMs); // delay before (c) message

    SDL_TimerID _timer = SDL_AddTimer(insertKeystrokesDelayMs, (SDL_TimerCallback)_process_event, NULL);
}

void eventprogrammer_start(Uint16 desiredFrameMs) {
    _frameMs = desiredFrameMs;
    _delayBetweenEventsMs = _frameMs * DELAY_BETWEEN_EVENTS_IN_FRAMES;

    SDL_Event e;
    Uint16 index = 0;

    // type "j" to get "LOAD"
    e.type = SDL_KEYDOWN;
    e.key.keysym.sym = SDLK_j;
    _loadQuoteQuoteEvents[index++] = e;
    e.type = SDL_KEYUP;
    e.key.keysym.sym = SDLK_j;
    _loadQuoteQuoteEvents[index++] = e;

    // hold shift
    e.type = SDL_KEYDOWN;
    e.key.keysym.sym = SDLK_RSHIFT;
    _loadQuoteQuoteEvents[index++] = e;

    // press "p" to get first double quote
    e.type = SDL_KEYDOWN;
    e.key.keysym.sym = SDLK_p;
    _loadQuoteQuoteEvents[index++] = e;
    e.type = SDL_KEYUP;
    e.key.keysym.sym = SDLK_p;
    _loadQuoteQuoteEvents[index++] = e;

    // press "p" again to get second double quote
    e.type = SDL_KEYDOWN;
    e.key.keysym.sym = SDLK_p;
    _loadQuoteQuoteEvents[index++] = e;
    e.type = SDL_KEYUP;
    e.key.keysym.sym = SDLK_p;
    _loadQuoteQuoteEvents[index++] = e;

    // release shift
    e.type = SDL_KEYUP;
    e.key.keysym.sym = SDLK_RSHIFT;
    _loadQuoteQuoteEvents[index++] = e;

    // press ENTER
    e.type = SDL_KEYDOWN;
    e.key.keysym.sym = SDLK_RETURN;
    _loadQuoteQuoteEvents[index++] = e;
    e.type = SDL_KEYUP;
    e.key.keysym.sym = SDLK_RETURN;
    _loadQuoteQuoteEvents[index++] = e;

    _loadQuoteQuoteNumEvents = index;

    _isInitialized = 1;
}

void eventprogrammer_destroy() {
    if (!_isInitialized) {
        return;
    }
    SDL_RemoveTimer(_timer);
    _isInitialized = 0;
}