#include "Timing.h"

Uint32 _timing_tstates_per_scanline__default;
Uint64 _timing_tstates_per_video_frame__default;

Uint32 _timing_tstates_per_scanline;
Uint64 _timing_tstates_per_video_frame;

Uint16 _timing_milliseconds_per_video_frame;
Uint8 _timing_tstate_factor;

Uint8 _timing_is_ready_to_destroy = 0;

Uint32 _last_scanline_tstate_overflow = 0;
Uint32 _per_scanline_adjustment_countdown_after_factor_change = 0;

// experimental
#define PER_SCANLINE_TSTATE_COMPENSATION_FACTOR 1.5f

Uint32 timing_get_states_per_scanline() {
	// we make this scan line pay the tstates borrowed by the previous
	// a closer to 100% means less sound desynchronization
	Uint32 compensatedOverflow = (Uint32)(PER_SCANLINE_TSTATE_COMPENSATION_FACTOR * (double)_last_scanline_tstate_overflow);
	return _timing_tstates_per_scanline - compensatedOverflow;
}

void _timing_handle_adjustment_after_tstate_factor_change(Uint32 actualTstates) {
	if (_per_scanline_adjustment_countdown_after_factor_change > 0) {
		// countdown after a tstate factor change
		_per_scanline_adjustment_countdown_after_factor_change--;
		_last_scanline_tstate_overflow = 0;
		return;
	}

	if (actualTstates <= _timing_tstates_per_scanline) {
		_last_scanline_tstate_overflow = 0;
		return;
	}

	// we had an overflow
	_last_scanline_tstate_overflow = actualTstates - _timing_tstates_per_scanline;
}

void timing_report_actual_states_last_scanline(Uint32 actualTstates) {
	_timing_handle_adjustment_after_tstate_factor_change(actualTstates);
}

Uint64 timing_get_states_per_video_frame() {
	return _timing_tstates_per_video_frame;
}

Uint16 timing_get_milliseconds_per_video_frame() {
	return _timing_milliseconds_per_video_frame;
}

void _timing_apply_tstate_factor__worker(factor) {
	_last_scanline_tstate_overflow = 0;
	_per_scanline_adjustment_countdown_after_factor_change = 5;

	_timing_tstate_factor = factor;
	_timing_tstates_per_scanline = _timing_tstates_per_scanline__default * _timing_tstate_factor;
	_timing_tstates_per_video_frame = _timing_tstates_per_video_frame__default * _timing_tstate_factor;
}

void timing_apply_tstate_factor(Uint8 factor) {
	if (_timing_is_ready_to_destroy) {
		return;
	}

	_timing_apply_tstate_factor__worker(factor);
}

Uint8 timing_is_faster() {
	return _timing_tstate_factor > 1;
}

void timing_prepare_destroy() {
	_timing_is_ready_to_destroy = 1;
	_timing_apply_tstate_factor__worker(1);
}

void timing_start(Uint16 millisecondsPerVideoFrame) {
	// 24 tstates for left border
	// 128 tstates for paper
	// 24 tstates for right border
	// 48 tstates for horizontal retrace
	_timing_tstates_per_scanline__default = 224;

	_timing_milliseconds_per_video_frame = millisecondsPerVideoFrame;
	_timing_tstates_per_video_frame__default = (Uint64)(SPECTRUM_SCREEN_HEIGHT) * (_timing_tstates_per_scanline__default);

	timing_apply_tstate_factor(1);
}

void timing_destroy() {

}
